#include <stdio.h>
#include <stdlib.h>
#include <poll.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <signal.h>
#include <termios.h>
#include <time.h>
#include <ctype.h>
#include <stdbool.h>

#define PORT "9999"  // the port users will be connecting to

#define BACKLOG 10   // how many pending connections queue will hold

// get sockaddr, IPv4 or IPv6:
void *get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}

	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

int main(void) {
	struct termios termios_p;
	if (tcgetattr(STDIN_FILENO, &termios_p) < 0) {
		fprintf(stderr, "failed to get attr: %d, %s\n", STDIN_FILENO, strerror(errno));
		exit(EXIT_FAILURE);
	}
	termios_p.c_lflag &= ~(ECHO | ICANON);
	if (tcsetattr(STDIN_FILENO, TCSANOW, &termios_p) < 0) {
		fprintf(stderr, "failed to set attr: %d, %s\n", STDIN_FILENO, strerror(errno));
		exit(EXIT_FAILURE);
	}
	int sockfd, new_fd, new_fd_1;  // listen on sock_fd, new connection on new_fd
	struct addrinfo hints, *servinfo, *p;
	struct sockaddr_storage their_addr; // connector's address information
	socklen_t sin_size;
	struct sigaction sa;
	int yes=1;
	char s[INET6_ADDRSTRLEN];
	int rv;

	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE; // use my IP

	if ((rv = getaddrinfo(NULL, PORT, &hints, &servinfo)) != 0) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
		return 1;
	}

	// loop through all the results and bind to the first we can
	for (p = servinfo; p != NULL; p = p->ai_next) {
		if ((sockfd = socket(p->ai_family, p->ai_socktype,
				p->ai_protocol)) == -1) {
			perror("server: socket");
			continue;
		}

		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes,
				sizeof(int)) == -1) {
			perror("setsockopt");
			exit(1);
		}

		if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
			close(sockfd);
			perror("server: bind");
			continue;
		}

		break;
	}

	freeaddrinfo(servinfo); // all done with this structure

	if (p == NULL)  {
		fprintf(stderr, "server: failed to bind\n");
		exit(1);
	}

	if (listen(sockfd, BACKLOG) == -1) {
		perror("listen");
		exit(1);
	}

	printf("server: waiting for connection...\n");

	sin_size = sizeof their_addr;
	new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
	if (new_fd == -1) {
		perror("accept");
	}

	inet_ntop(their_addr.ss_family,
		get_in_addr((struct sockaddr *)&their_addr),
		s, sizeof s);
	printf("server: got connection from %s\n", s);
	new_fd_1 = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
	if (new_fd_1 == -1) {
		perror("accept");
	}

	inet_ntop(their_addr.ss_family,
		get_in_addr((struct sockaddr *)&their_addr),
		s, sizeof s);
	printf("server: got connection from %s\n", s);
	struct pollfd psfd, psfd1;
	psfd.fd = new_fd;
	psfd.events = POLLIN;
	psfd1.fd = new_fd_1;
	psfd1.events = POLLIN;
	char lineptr[BUFSIZ] = { 0, };
	ssize_t n = 0;
	const struct timespec tspec = { .tv_sec = 0, .tv_nsec = 16666666 };
	while (1) {
		poll(&psfd, 1, 0);
		if (psfd.revents & POLLIN) {
			char buf[BUFSIZ];
			ssize_t byts = read(new_fd, buf, BUFSIZ);
			write(new_fd, buf, byts);
			write(new_fd_1, buf, byts);
		}
		poll(&psfd1, 1, 0);
		if (psfd1.revents & POLLIN) {
			char buf[BUFSIZ];
			ssize_t byts = read(new_fd_1, buf, BUFSIZ);
			write(new_fd, buf, byts);
			write(new_fd_1, buf, byts);
		}
		nanosleep(&tspec, NULL);
	}
}
